
from argparse import Namespace
import io
import os
import pathlib
import shutil

import yaml

from invariant_client import aws_pruner, zip_util


def use_aws_pruner(tempdir: os.PathLike, args: Namespace, aws_pruner_debug_out: os.PathLike | None) -> io.BytesIO:
    # Use zipfile to inspect 'bytes', test if invariant/aws_pruner.yaml is present
    try:
        if pathlib.Path(tempdir, 'invariant', 'aws_pruner.yaml').exists():
            with open(pathlib.Path(tempdir, 'invariant', 'aws_pruner.yaml'), 'r') as f:
                pruner_config = aws_pruner.UserConfig.from_yaml(f.read())
        else:
            pruner_config = aws_pruner.UserConfig.from_dict({})
        apply_pruner = not getattr(args, 'no_aws_pruner') and pruner_config.enabled
        if apply_pruner:
            print("Pruning EC2 instances and network interfaces")
        else:
            print("Pruner dry run: pruning EC2 instances and network interfaces")

        # Discover all Reservations.json files and run the pruner on each (modify in place)
        for reservations_path in pathlib.Path(tempdir, 'aws_configs').rglob("Reservations.json"):
            path = pathlib.Path(reservations_path).parent
            print(f"Pruner: processing {path.relative_to(tempdir)}")
            aws_prune_tool = aws_pruner.AwsPruneTool(
                path,
                pruner_config.prune_level,
                pruner_config.filter_exclude,
                pruner_config.filter_include,
                pruner_config.group_by,
            )
            aws_prune_tool.load()
            aws_prune_tool.execute()
            aws_prune_tool.write()
        if aws_pruner_debug_out is not None:
            # Copy the contents of the aws_config directory in tempdir to aws_pruner_debug_out
            debug_out_path = pathlib.Path(aws_pruner_debug_out)
            if debug_out_path.is_dir():
                # remove contents
                shutil.rmtree(debug_out_path)
            print(f"Pruner writing debug output to {debug_out_path} ...")
            shutil.copytree(pathlib.Path(tempdir, 'aws_configs'), debug_out_path)
        return apply_pruner

    except yaml.YAMLError as e:
        print(f"Error parsing YAML: {e}")
        return None
    except ValueError as e:
        print(f"Error loading invariant/aws_pruner.yaml: {e}")
        return None


from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.basic_auth_login_method import BasicAuthLoginMethod
    from ..models.oidc_login_method import OIDCLoginMethod


T = TypeVar("T", bound="CreateManagedUserRequest")


@_attrs_define
class CreateManagedUserRequest:
    """
    Attributes:
        type_ (Literal['managed']):
        email (str):
        allowed_methods (Union[None, Unset, list[Union['BasicAuthLoginMethod', 'OIDCLoginMethod']]]):
        send_invite (Union[Unset, bool]):  Default: False.
        use_setup_code (Union[Unset, bool]):  Default: False.
        is_superuser (Union[Unset, bool]):  Default: False.
    """

    type_: Literal["managed"]
    email: str
    allowed_methods: Union[
        None, Unset, list[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]]
    ] = UNSET
    send_invite: Union[Unset, bool] = False
    use_setup_code: Union[Unset, bool] = False
    is_superuser: Union[Unset, bool] = False
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        type_ = self.type_

        email = self.email

        allowed_methods: Union[None, Unset, list[dict[str, Any]]]
        if isinstance(self.allowed_methods, Unset):
            allowed_methods = UNSET
        elif isinstance(self.allowed_methods, list):
            allowed_methods = []
            for allowed_methods_type_0_item_data in self.allowed_methods:
                allowed_methods_type_0_item: dict[str, Any]
                if isinstance(allowed_methods_type_0_item_data, BasicAuthLoginMethod):
                    allowed_methods_type_0_item = (
                        allowed_methods_type_0_item_data.to_dict()
                    )
                else:
                    allowed_methods_type_0_item = (
                        allowed_methods_type_0_item_data.to_dict()
                    )

                allowed_methods.append(allowed_methods_type_0_item)

        else:
            allowed_methods = self.allowed_methods

        send_invite = self.send_invite

        use_setup_code = self.use_setup_code

        is_superuser = self.is_superuser

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type_,
                "email": email,
            }
        )
        if allowed_methods is not UNSET:
            field_dict["allowed_methods"] = allowed_methods
        if send_invite is not UNSET:
            field_dict["send_invite"] = send_invite
        if use_setup_code is not UNSET:
            field_dict["use_setup_code"] = use_setup_code
        if is_superuser is not UNSET:
            field_dict["is_superuser"] = is_superuser

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.basic_auth_login_method import BasicAuthLoginMethod
        from ..models.oidc_login_method import OIDCLoginMethod

        d = dict(src_dict)
        type_ = cast(Literal["managed"], d.pop("type"))
        if type_ != "managed":
            raise ValueError(f"type must match const 'managed', got '{type_}'")

        email = d.pop("email")

        def _parse_allowed_methods(
            data: object,
        ) -> Union[None, Unset, list[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                allowed_methods_type_0 = []
                _allowed_methods_type_0 = data
                for allowed_methods_type_0_item_data in _allowed_methods_type_0:

                    def _parse_allowed_methods_type_0_item(
                        data: object,
                    ) -> Union["BasicAuthLoginMethod", "OIDCLoginMethod"]:
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            allowed_methods_type_0_item_type_0 = (
                                BasicAuthLoginMethod.from_dict(data)
                            )

                            return allowed_methods_type_0_item_type_0
                        except:  # noqa: E722
                            pass
                        if not isinstance(data, dict):
                            raise TypeError()
                        allowed_methods_type_0_item_type_1 = OIDCLoginMethod.from_dict(
                            data
                        )

                        return allowed_methods_type_0_item_type_1

                    allowed_methods_type_0_item = _parse_allowed_methods_type_0_item(
                        allowed_methods_type_0_item_data
                    )

                    allowed_methods_type_0.append(allowed_methods_type_0_item)

                return allowed_methods_type_0
            except:  # noqa: E722
                pass
            return cast(
                Union[
                    None, Unset, list[Union["BasicAuthLoginMethod", "OIDCLoginMethod"]]
                ],
                data,
            )

        allowed_methods = _parse_allowed_methods(d.pop("allowed_methods", UNSET))

        send_invite = d.pop("send_invite", UNSET)

        use_setup_code = d.pop("use_setup_code", UNSET)

        is_superuser = d.pop("is_superuser", UNSET)

        create_managed_user_request = cls(
            type_=type_,
            email=email,
            allowed_methods=allowed_methods,
            send_invite=send_invite,
            use_setup_code=use_setup_code,
            is_superuser=is_superuser,
        )

        create_managed_user_request.additional_properties = d
        return create_managed_user_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    Any,
    Literal,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="CreateServiceAccountRequest")


@_attrs_define
class CreateServiceAccountRequest:
    """
    Attributes:
        type_ (Literal['service_account']):
        sa_name (str):
        sa_comment (Union[None, Unset, str]):
        is_superuser (Union[Unset, bool]):  Default: False.
    """

    type_: Literal["service_account"]
    sa_name: str
    sa_comment: Union[None, Unset, str] = UNSET
    is_superuser: Union[Unset, bool] = False
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        type_ = self.type_

        sa_name = self.sa_name

        sa_comment: Union[None, Unset, str]
        if isinstance(self.sa_comment, Unset):
            sa_comment = UNSET
        else:
            sa_comment = self.sa_comment

        is_superuser = self.is_superuser

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type_,
                "sa_name": sa_name,
            }
        )
        if sa_comment is not UNSET:
            field_dict["sa_comment"] = sa_comment
        if is_superuser is not UNSET:
            field_dict["is_superuser"] = is_superuser

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        type_ = cast(Literal["service_account"], d.pop("type"))
        if type_ != "service_account":
            raise ValueError(f"type must match const 'service_account', got '{type_}'")

        sa_name = d.pop("sa_name")

        def _parse_sa_comment(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        sa_comment = _parse_sa_comment(d.pop("sa_comment", UNSET))

        is_superuser = d.pop("is_superuser", UNSET)

        create_service_account_request = cls(
            type_=type_,
            sa_name=sa_name,
            sa_comment=sa_comment,
            is_superuser=is_superuser,
        )

        create_service_account_request.additional_properties = d
        return create_service_account_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.policy_file_result_rule import PolicyFileResultRule


T = TypeVar("T", bound="PolicyFileResultPolicy")


@_attrs_define
class PolicyFileResultPolicy:
    """
    Attributes:
        name (str):
        rules (list['PolicyFileResultRule']):
        policy_errors (Union[None, Unset, list[str]]):
        comment (Union[None, Unset, str]):
        owner (Union[None, Unset, str]):
    """

    name: str
    rules: list["PolicyFileResultRule"]
    policy_errors: Union[None, Unset, list[str]] = UNSET
    comment: Union[None, Unset, str] = UNSET
    owner: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        rules = []
        for rules_item_data in self.rules:
            rules_item = rules_item_data.to_dict()
            rules.append(rules_item)

        policy_errors: Union[None, Unset, list[str]]
        if isinstance(self.policy_errors, Unset):
            policy_errors = UNSET
        elif isinstance(self.policy_errors, list):
            policy_errors = self.policy_errors

        else:
            policy_errors = self.policy_errors

        comment: Union[None, Unset, str]
        if isinstance(self.comment, Unset):
            comment = UNSET
        else:
            comment = self.comment

        owner: Union[None, Unset, str]
        if isinstance(self.owner, Unset):
            owner = UNSET
        else:
            owner = self.owner

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "rules": rules,
            }
        )
        if policy_errors is not UNSET:
            field_dict["policy_errors"] = policy_errors
        if comment is not UNSET:
            field_dict["comment"] = comment
        if owner is not UNSET:
            field_dict["owner"] = owner

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.policy_file_result_rule import PolicyFileResultRule

        d = dict(src_dict)
        name = d.pop("name")

        rules = []
        _rules = d.pop("rules")
        for rules_item_data in _rules:
            rules_item = PolicyFileResultRule.from_dict(rules_item_data)

            rules.append(rules_item)

        def _parse_policy_errors(data: object) -> Union[None, Unset, list[str]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                policy_errors_type_0 = cast(list[str], data)

                return policy_errors_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[str]], data)

        policy_errors = _parse_policy_errors(d.pop("policy_errors", UNSET))

        def _parse_comment(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        comment = _parse_comment(d.pop("comment", UNSET))

        def _parse_owner(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        owner = _parse_owner(d.pop("owner", UNSET))

        policy_file_result_policy = cls(
            name=name,
            rules=rules,
            policy_errors=policy_errors,
            comment=comment,
            owner=owner,
        )

        policy_file_result_policy.additional_properties = d
        return policy_file_result_policy

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="IntegrationDataSlackAppInstallationData")


@_attrs_define
class IntegrationDataSlackAppInstallationData:
    """
    Attributes:
        bot_token (str):
        team_id (str):
        team_name (str):
        user_creator (UUID):
    """

    bot_token: str
    team_id: str
    team_name: str
    user_creator: UUID
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        bot_token = self.bot_token

        team_id = self.team_id

        team_name = self.team_name

        user_creator = str(self.user_creator)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "bot_token": bot_token,
                "team_id": team_id,
                "team_name": team_name,
                "user_creator": user_creator,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        bot_token = d.pop("bot_token")

        team_id = d.pop("team_id")

        team_name = d.pop("team_name")

        user_creator = UUID(d.pop("user_creator"))

        integration_data_slack_app_installation_data = cls(
            bot_token=bot_token,
            team_id=team_id,
            team_name=team_name,
            user_creator=user_creator,
        )

        integration_data_slack_app_installation_data.additional_properties = d
        return integration_data_slack_app_installation_data

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.list_snapshots_response import ListSnapshotsResponse
    from ..models.network import Network


T = TypeVar("T", bound="NetworkWithExtras")


@_attrs_define
class NetworkWithExtras:
    """
    Attributes:
        network (Network):
        current_snapshot (Union['ListSnapshotsResponse', None, Unset]):
    """

    network: "Network"
    current_snapshot: Union["ListSnapshotsResponse", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.list_snapshots_response import ListSnapshotsResponse

        network = self.network.to_dict()

        current_snapshot: Union[None, Unset, dict[str, Any]]
        if isinstance(self.current_snapshot, Unset):
            current_snapshot = UNSET
        elif isinstance(self.current_snapshot, ListSnapshotsResponse):
            current_snapshot = self.current_snapshot.to_dict()
        else:
            current_snapshot = self.current_snapshot

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "network": network,
            }
        )
        if current_snapshot is not UNSET:
            field_dict["current_snapshot"] = current_snapshot

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.list_snapshots_response import ListSnapshotsResponse
        from ..models.network import Network

        d = dict(src_dict)
        network = Network.from_dict(d.pop("network"))

        def _parse_current_snapshot(
            data: object,
        ) -> Union["ListSnapshotsResponse", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                current_snapshot_type_0 = ListSnapshotsResponse.from_dict(data)

                return current_snapshot_type_0
            except:  # noqa: E722
                pass
            return cast(Union["ListSnapshotsResponse", None, Unset], data)

        current_snapshot = _parse_current_snapshot(d.pop("current_snapshot", UNSET))

        network_with_extras = cls(
            network=network,
            current_snapshot=current_snapshot,
        )

        network_with_extras.additional_properties = d
        return network_with_extras

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class BuiltInRole(str, Enum):
    EVALUATOR = "evaluator"
    NETWORK_MANAGER = "network_manager"
    UPLOAD_ONLY = "upload_only"
    WORKSPACE_MANAGER = "workspace_manager"

    def __str__(self) -> str:
        return str(self.value)



from pydantic import (
    BaseModel,
    FilePath,
    ValidationError,
    Field,
    model_validator,
    FieldValidationInfo,
    field_validator,
    AfterValidator,
    SecretStr,
)
from typing import Literal, Optional, List, Union, Any
from enum import Enum
import yaml
from pathlib import Path
from typing import Annotated
from invariant_client.loaders import load


class SourceKind(str, Enum):
    LIBRENMS = "librenms"
    AWS = "aws"


class BaseSourceConfig(BaseModel):
    name: str
    kind: SourceKind  # Pydantic handles SourceKind enum correctly for discrimination


class LibreNMSConfig(BaseSourceConfig):
    # Using the enum member for Literal is fine for str enums
    kind: Literal[SourceKind.LIBRENMS]
    hostname: str
    device_group: str
    api_key: SecretStr
    ssh_key: Optional[SecretStr] = None
    ssh_key_path: Optional[FilePath] = None
    ssh_user: Optional[str] = None

    @model_validator(mode="after")
    def _check_ssh_key_or_path_exclusive(self) -> "LibreNMSConfig":
        ssh_key_present = self.ssh_key is not None
        ssh_key_path_present = self.ssh_key_path is not None

        if ssh_key_present and ssh_key_path_present:
            raise ValueError(
                f"Error in source {self.name}: Fields 'ssh_key' and 'ssh_key_path' are mutually exclusive. Provide only one."
            )
        if not ssh_key_present and not ssh_key_path_present:
            raise ValueError(
                f"Error in source {self.name}: Exactly one of 'ssh_key' or 'ssh_key_path' is required."
            )
        return self

    @model_validator(mode="after")
    def _resolve_secrets_in_model(self) -> "LibreNMSConfig":
        if self.api_key:
            if "://" in self.api_key.get_secret_value():
                self.api_key = SecretStr(load(self.api_key.get_secret_value()))
        if self.ssh_key:
            if "://" in self.ssh_key.get_secret_value():
                self.ssh_key = SecretStr(load(self.ssh_key.get_secret_value()))
        return self


class AWSConfig(BaseSourceConfig):
    kind: Literal[SourceKind.AWS]  # Using the enum member for Literal is fine
    profile: Optional[str] = None
    role: Optional[str] = None
    regions: List[str] = []
    accounts: List[str] = []
    ignore_accounts: List[str] = []
    skip_resources: List[str] = []


Sources = Annotated[Union[LibreNMSConfig, AWSConfig], Field(discriminator="kind")]


class TopLevelConfig(BaseModel):
    model_config = {
        "deprecated": {
            "output_path": "The 'output_path' field is deprecated and will be removed in a future release."
        }
    }
    sources: List[Sources]  # Pydantic uses 'kind' for discrimination
    output_path: Optional[Path] = None


def validate_yaml(yaml_file: dict, context: Optional[dict] = None) -> TopLevelConfig:
    try:
        return TopLevelConfig.model_validate(yaml_file)
    except yaml.YAMLError as e:
        raise ValueError(f"Invalid YAML: {e}")
    except ValidationError as e:
        raise ValueError(f"Configuration validation error:\n{e}") from e



from collections.abc import Mapping
from typing import (
    Any,
    Literal,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.invariant_plan import InvariantPlan
from ..types import UNSET, Unset

T = TypeVar("T", bound="SetupCheckoutRequest")


@_attrs_define
class SetupCheckoutRequest:
    """
    Attributes:
        plan (InvariantPlan): Organization subscription plans.
        type_ (Union[Literal['plan_change'], Unset]):  Default: 'plan_change'.
        capacity (Union[None, Unset, int]):
    """

    plan: InvariantPlan
    type_: Union[Literal["plan_change"], Unset] = "plan_change"
    capacity: Union[None, Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        plan = self.plan.value

        type_ = self.type_

        capacity: Union[None, Unset, int]
        if isinstance(self.capacity, Unset):
            capacity = UNSET
        else:
            capacity = self.capacity

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "plan": plan,
            }
        )
        if type_ is not UNSET:
            field_dict["type"] = type_
        if capacity is not UNSET:
            field_dict["capacity"] = capacity

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        plan = InvariantPlan(d.pop("plan"))

        type_ = cast(Union[Literal["plan_change"], Unset], d.pop("type", UNSET))
        if type_ != "plan_change" and not isinstance(type_, Unset):
            raise ValueError(f"type must match const 'plan_change', got '{type_}'")

        def _parse_capacity(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        capacity = _parse_capacity(d.pop("capacity", UNSET))

        setup_checkout_request = cls(
            plan=plan,
            type_=type_,
            capacity=capacity,
        )

        setup_checkout_request.additional_properties = d
        return setup_checkout_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    Any,
    Literal,
    TypeVar,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="ModifyAllowOutboundInvitationsRequest")


@_attrs_define
class ModifyAllowOutboundInvitationsRequest:
    """
    Attributes:
        policy_key (Literal['allow_outbound_invitations']):
        value (bool):
    """

    policy_key: Literal["allow_outbound_invitations"]
    value: bool
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        policy_key = self.policy_key

        value = self.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "policy_key": policy_key,
                "value": value,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        policy_key = cast(Literal["allow_outbound_invitations"], d.pop("policy_key"))
        if policy_key != "allow_outbound_invitations":
            raise ValueError(
                f"policy_key must match const 'allow_outbound_invitations', got '{policy_key}'"
            )

        value = d.pop("value")

        modify_allow_outbound_invitations_request = cls(
            policy_key=policy_key,
            value=value,
        )

        modify_allow_outbound_invitations_request.additional_properties = d
        return modify_allow_outbound_invitations_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.user import User


T = TypeVar("T", bound="NetworkUserAccessActorV1")


@_attrs_define
class NetworkUserAccessActorV1:
    """
    Attributes:
        uuid (UUID):
        user (User):
        source_address (list[Union[int, str]]):
        type_ (Union[Literal['network_user'], Unset]):  Default: 'network_user'.
    """

    uuid: UUID
    user: "User"
    source_address: list[Union[int, str]]
    type_: Union[Literal["network_user"], Unset] = "network_user"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid = str(self.uuid)

        user = self.user.to_dict()

        source_address = []
        for componentsschemas_address_item_data in self.source_address:
            componentsschemas_address_item: Union[int, str]
            componentsschemas_address_item = componentsschemas_address_item_data
            source_address.append(componentsschemas_address_item)

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "uuid": uuid,
                "user": user,
                "source_address": source_address,
            }
        )
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.user import User

        d = dict(src_dict)
        uuid = UUID(d.pop("uuid"))

        user = User.from_dict(d.pop("user"))

        source_address = []
        _source_address = d.pop("source_address")
        for componentsschemas_address_item_data in _source_address:

            def _parse_componentsschemas_address_item(data: object) -> Union[int, str]:
                return cast(Union[int, str], data)

            componentsschemas_address_item = _parse_componentsschemas_address_item(
                componentsschemas_address_item_data
            )

            source_address.append(componentsschemas_address_item)

        type_ = cast(Union[Literal["network_user"], Unset], d.pop("type", UNSET))
        if type_ != "network_user" and not isinstance(type_, Unset):
            raise ValueError(f"type must match const 'network_user', got '{type_}'")

        network_user_access_actor_v1 = cls(
            uuid=uuid,
            user=user,
            source_address=source_address,
            type_=type_,
        )

        network_user_access_actor_v1.additional_properties = d
        return network_user_access_actor_v1

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    Any,
    Literal,
    TypeVar,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="PasswordResetPINChallenge")


@_attrs_define
class PasswordResetPINChallenge:
    """The user must prove control of the email address by PIN before proceeding to reset the password.

    Attributes:
        type_ (Literal['reset_pin']):
    """

    type_: Literal["reset_pin"]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type_,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        type_ = cast(Literal["reset_pin"], d.pop("type"))
        if type_ != "reset_pin":
            raise ValueError(f"type must match const 'reset_pin', got '{type_}'")

        password_reset_pin_challenge = cls(
            type_=type_,
        )

        password_reset_pin_challenge.additional_properties = d
        return password_reset_pin_challenge

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import logging
import os
import pathlib
import re
import sys
import yaml

logger = logging.getLogger(__name__)


class RulePickError(Exception):
    """Base exception for RulePick errors."""

class SelectorError(RulePickError):
    """Base for errors related to selector parsing or matching."""

class SelectorFormatError(SelectorError):
    """Raised when the selector format is invalid."""

class PolicyNotFoundError(SelectorError):
    """Raised when the specified policy is not found."""

class RuleNotFoundError(SelectorError):
    """Raised when the specified rule (by index) is not found."""

class InvalidPolicyFileFormatError(RulePickError):
    """Raised when the policy file format is invalid."""


# policy_name[index] or policy_index[index]
SELECTOR_INDEX_REGEX = re.compile(r"^(?P<policy_name>[^\[]+)\[(?P<index>\d+)\]$")
# just a number
SELECTOR_GLOBAL_INDEX_REGEX = re.compile(r"^(?P<index>\d+)$")


def pick_rule(target_policy_file: str | os.PathLike, selector: str | None) -> str:
    """
    Selects a specific rule from an access policy YAML file based on a selector
    and generates a new YAML string containing only that rule within its
    original parent policy structure.

    Prints user-facing error messages to stderr before raising exceptions
    for errors like file not found, bad YAML, invalid selector, or rule/policy
    not found.

    Supported Selector Formats:
      - 'policy_name[index]': Selects rule by index within a named policy.
      - 'policy_index[index]': Selects rule by index within a policy at a given index.
      - 'index': Selects rule by its global index across all policies in the file.

    Args:
        target_yaml_file: Path to the input YAML file.
        selector: String specifying the rule to select.

    Returns:
        A string containing the YAML representation of the selected rule within its policy.

    Raises:
        FileNotFoundError: If the target_yaml_file does not exist.
        yaml.YAMLError: If the target_yaml_file is not valid YAML.
        InvalidPolicyFileFormatError: If the YAML structure is incorrect.
        SelectorFormatError: If the selector string format is invalid or ambiguous.
        PolicyNotFoundError: If the policy specified in the selector is not found.
        RuleNotFoundError: If the rule specified in the selector is not found.
        RulePickError: For other generic errors during processing.
    """
    target_path = pathlib.Path(target_policy_file)
    loaded_data = _load_yaml_file(target_path)
    policies = _validate_and_get_policies(loaded_data, target_path)

    selected_policy = None
    selected_rule = None

    if selector is None:
        selected_policy, selected_rule = _find_by_global_rule_index(
            policies, 0, target_path, no_selector=True
        )
    else:
        # --- Determine selector type and find rule/policy ---
        index_match = SELECTOR_INDEX_REGEX.match(selector)
        global_index_match = SELECTOR_GLOBAL_INDEX_REGEX.match(selector)

        if index_match:
            policy_selector = index_match.group('policy_name')
            rule_index = int(index_match.group('index'))
            selected_policy, selected_rule = _find_by_policy_and_rule_index(
                policies, policy_selector, rule_index, target_path
            )
        elif global_index_match:
            rule_index = int(global_index_match.group('index'))
            selected_policy, selected_rule = _find_by_global_rule_index(
                policies, rule_index, target_path
            )
        else:
            msg = f"Invalid selector format: '{selector}'. Supported formats: 'policy[index]' or 'index'."
            print(f"Error: {msg}", file=sys.stderr)
            raise SelectorFormatError(msg)

    # --- Reconstruct Output ---
    output_policy = {}
    # Copy top-level keys/values from the original policy, excluding 'rules'
    for key, value in selected_policy.items():
        if key != 'rules':
            output_policy[key] = value # Shallow copy of values is sufficient

    # Add the single selected rule (reference is fine)
    output_policy['rules'] = [selected_rule]

    # Wrap in the top-level structure
    output_data = {'access-policy': [output_policy]}

    # Write YAML
    return yaml.dump(output_data, sort_keys=False, default_flow_style=False, indent=2)

def _load_yaml_file(target_path: pathlib.Path) -> dict:
    """Loads and performs basic validation on the YAML file."""
    try:
        with open(target_path, 'r', encoding='utf-8') as f:
            loaded_data = yaml.safe_load(f)
    except FileNotFoundError:
        print(f"Error: Policy file not found at '{target_path}'", file=sys.stderr)
        raise
    except yaml.YAMLError as e:
        print(f"Error: Failed to parse YAML file '{target_path}': {e}", file=sys.stderr)
        raise
    except OSError as e: # Other potential OS errors
        print(f"Error: Could not read file '{target_path}': {e}", file=sys.stderr)
        raise RulePickError(f"Could not read file '{target_path}'") from e

    if not isinstance(loaded_data, dict):
        msg = f"Invalid policy file format: Top level must be a dictionary (object) in '{target_path}'"
        print(f"Error: {msg}", file=sys.stderr)
        raise InvalidPolicyFileFormatError(msg)

    return loaded_data

def _validate_and_get_policies(loaded_data: dict, target_path: pathlib.Path) -> list:
    """Validates the 'access-policy' key and returns the list of policies."""
    if 'access-policy' not in loaded_data:
        msg = f"Invalid policy file format: missing top-level 'access-policy' key in '{target_path}'"
        print(f"Error: {msg}", file=sys.stderr)
        raise InvalidPolicyFileFormatError(msg)

    policies = loaded_data.get('access-policy')
    if not isinstance(policies, list):
        msg = f"Invalid policy file format: 'access-policy' should contain a list in '{target_path}'"
        print(f"Error: {msg}", file=sys.stderr)
        raise InvalidPolicyFileFormatError(msg)
    return policies

def _find_by_policy_and_rule_index(policies: list, policy_selector: str, rule_index: int, target_path: pathlib.Path) -> tuple[dict, dict]:
    """Finds policy by name or index, then rule by index."""
    found_policy = None
    # First, try to find a policy by name
    for policy in policies:
        if isinstance(policy, dict) and policy.get('name') == policy_selector:
            found_policy = policy
            break

    # If no policy is found by name and the selector is a number, try to use it as an index
    if found_policy is None and policy_selector.isdigit():
        policy_index = int(policy_selector)
        if 0 <= policy_index < len(policies):
            found_policy = policies[policy_index]
        else:
            msg = f"Policy index {policy_index} out of bounds (has {len(policies)} policies) in '{target_path}'"
            print(f"Error: {msg}", file=sys.stderr)
            raise PolicyNotFoundError(msg)

    if found_policy is None:
        msg = f"Policy '{policy_selector}' not found in '{target_path}'"
        print(f"Error: {msg}", file=sys.stderr)
        raise PolicyNotFoundError(msg)

    rules = found_policy.get('rules')
    if not isinstance(rules, list):
        policy_identifier = found_policy.get('name', f"at index {policies.index(found_policy)}")
        msg = f"Policy '{policy_identifier}' is missing or has an invalid 'rules' list in '{target_path}'"
        print(f"Error: {msg}", file=sys.stderr)
        raise InvalidPolicyFileFormatError(msg)

    if not (0 <= rule_index < len(rules)):
        policy_identifier = found_policy.get('name', f"at index {policies.index(found_policy)}")
        msg = f"Rule index {rule_index} out of bounds for policy '{policy_identifier}' (has {len(rules)} rules) in '{target_path}'"
        print(f"Error: {msg}", file=sys.stderr)
        raise RuleNotFoundError(msg)

    selected_rule = rules[rule_index]
    if not isinstance(selected_rule, dict):
        policy_identifier = found_policy.get('name', f"at index {policies.index(found_policy)}")
        msg = f"Invalid rule format at index {rule_index} in policy '{policy_identifier}' (expected dictionary) in '{target_path}'"
        print(f"Error: {msg}", file=sys.stderr)
        raise InvalidPolicyFileFormatError(msg)

    return found_policy, selected_rule

def _find_by_global_rule_index(
        policies: list,
        rule_index: int,
        target_path: pathlib.Path,
        no_selector: bool = False
    ) -> tuple[dict, dict]:
    """Finds a rule by its global index across all policies."""
    # Count all rules in policies (E.g. policies[0].rules)
    rule_count = sum(len(policy.get('rules', [])) for policy in policies if isinstance(policy, dict) and isinstance(policy.get('rules'), list))
    if no_selector and rule_count > 1:
        msg = f"Missing --rule <selector>. Rule selector required as multiple rules ({rule_count}) exist in '{target_path}'."
        raise ValueError(msg)
    current_rule_count = 0
    for policy in policies:
        if not isinstance(policy, dict):
            continue
        rules = policy.get('rules')
        if not isinstance(rules, list):
            continue

        if current_rule_count <= rule_index < current_rule_count + len(rules):
            local_rule_index = rule_index - current_rule_count
            selected_rule = rules[local_rule_index]
            if not isinstance(selected_rule, dict):
                msg = f"Invalid rule format at global index {rule_index} (expected dictionary) in '{target_path}'"
                print(f"Error: {msg}", file=sys.stderr)
                raise InvalidPolicyFileFormatError(msg)
            return policy, selected_rule

        current_rule_count += len(rules)

    msg = f"Global rule index {rule_index} out of bounds (total {current_rule_count} rules) in '{target_path}'"
    print(f"Error: {msg}", file=sys.stderr)
    raise RuleNotFoundError(msg)



import datetime
from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.built_in_role import BuiltInRole
from ..types import UNSET, Unset

T = TypeVar("T", bound="UserMetadata")


@_attrs_define
class UserMetadata:
    """
    Attributes:
        is_superuser (bool):
        needs_invite_link_version (Union[None, Unset, int]):
        invite_link_expires_at (Union[None, Unset, datetime.datetime]):
        is_service_account (Union[Unset, bool]):  Default: False.
        service_account_name (Union[None, Unset, str]):
        service_account_comment (Union[None, Unset, str]):
        role (Union[BuiltInRole, None, Unset]):
    """

    is_superuser: bool
    needs_invite_link_version: Union[None, Unset, int] = UNSET
    invite_link_expires_at: Union[None, Unset, datetime.datetime] = UNSET
    is_service_account: Union[Unset, bool] = False
    service_account_name: Union[None, Unset, str] = UNSET
    service_account_comment: Union[None, Unset, str] = UNSET
    role: Union[BuiltInRole, None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        is_superuser = self.is_superuser

        needs_invite_link_version: Union[None, Unset, int]
        if isinstance(self.needs_invite_link_version, Unset):
            needs_invite_link_version = UNSET
        else:
            needs_invite_link_version = self.needs_invite_link_version

        invite_link_expires_at: Union[None, Unset, str]
        if isinstance(self.invite_link_expires_at, Unset):
            invite_link_expires_at = UNSET
        elif isinstance(self.invite_link_expires_at, datetime.datetime):
            invite_link_expires_at = self.invite_link_expires_at.isoformat()
        else:
            invite_link_expires_at = self.invite_link_expires_at

        is_service_account = self.is_service_account

        service_account_name: Union[None, Unset, str]
        if isinstance(self.service_account_name, Unset):
            service_account_name = UNSET
        else:
            service_account_name = self.service_account_name

        service_account_comment: Union[None, Unset, str]
        if isinstance(self.service_account_comment, Unset):
            service_account_comment = UNSET
        else:
            service_account_comment = self.service_account_comment

        role: Union[None, Unset, str]
        if isinstance(self.role, Unset):
            role = UNSET
        elif isinstance(self.role, BuiltInRole):
            role = self.role.value
        else:
            role = self.role

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "is_superuser": is_superuser,
            }
        )
        if needs_invite_link_version is not UNSET:
            field_dict["needs_invite_link_version"] = needs_invite_link_version
        if invite_link_expires_at is not UNSET:
            field_dict["invite_link_expires_at"] = invite_link_expires_at
        if is_service_account is not UNSET:
            field_dict["is_service_account"] = is_service_account
        if service_account_name is not UNSET:
            field_dict["service_account_name"] = service_account_name
        if service_account_comment is not UNSET:
            field_dict["service_account_comment"] = service_account_comment
        if role is not UNSET:
            field_dict["role"] = role

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        is_superuser = d.pop("is_superuser")

        def _parse_needs_invite_link_version(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        needs_invite_link_version = _parse_needs_invite_link_version(
            d.pop("needs_invite_link_version", UNSET)
        )

        def _parse_invite_link_expires_at(
            data: object,
        ) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                invite_link_expires_at_type_0 = isoparse(data)

                return invite_link_expires_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        invite_link_expires_at = _parse_invite_link_expires_at(
            d.pop("invite_link_expires_at", UNSET)
        )

        is_service_account = d.pop("is_service_account", UNSET)

        def _parse_service_account_name(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        service_account_name = _parse_service_account_name(
            d.pop("service_account_name", UNSET)
        )

        def _parse_service_account_comment(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        service_account_comment = _parse_service_account_comment(
            d.pop("service_account_comment", UNSET)
        )

        def _parse_role(data: object) -> Union[BuiltInRole, None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                role_type_0 = BuiltInRole(data)

                return role_type_0
            except:  # noqa: E722
                pass
            return cast(Union[BuiltInRole, None, Unset], data)

        role = _parse_role(d.pop("role", UNSET))

        user_metadata = cls(
            is_superuser=is_superuser,
            needs_invite_link_version=needs_invite_link_version,
            invite_link_expires_at=invite_link_expires_at,
            is_service_account=is_service_account,
            service_account_name=service_account_name,
            service_account_comment=service_account_comment,
            role=role,
        )

        user_metadata.additional_properties = d
        return user_metadata

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

if TYPE_CHECKING:
    from ..models.access_session_context_type_0 import AccessSessionContextType0
    from ..models.network_ip_access_actor import NetworkIPAccessActor
    from ..models.network_user_access_actor import NetworkUserAccessActor
    from ..models.network_user_access_actor_v1 import NetworkUserAccessActorV1
    from ..models.system_access_actor import SystemAccessActor


T = TypeVar("T", bound="AccessSession")


@_attrs_define
class AccessSession:
    """Represents a data acess actor. This could represent an end user accessing data over the network or some internal
    system action.


        Attributes:
            organization_uuid (UUID):
            actor (Union['NetworkIPAccessActor', 'NetworkUserAccessActor', 'NetworkUserAccessActorV1',
                'SystemAccessActor']):
            context (Union['AccessSessionContextType0', None]):
            created_at (datetime.datetime):
            object_urn (str):
    """

    organization_uuid: UUID
    actor: Union[
        "NetworkIPAccessActor",
        "NetworkUserAccessActor",
        "NetworkUserAccessActorV1",
        "SystemAccessActor",
    ]
    context: Union["AccessSessionContextType0", None]
    created_at: datetime.datetime
    object_urn: str
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.access_session_context_type_0 import AccessSessionContextType0
        from ..models.network_ip_access_actor import NetworkIPAccessActor
        from ..models.network_user_access_actor import NetworkUserAccessActor
        from ..models.network_user_access_actor_v1 import NetworkUserAccessActorV1

        organization_uuid = str(self.organization_uuid)

        actor: dict[str, Any]
        if isinstance(self.actor, NetworkUserAccessActor):
            actor = self.actor.to_dict()
        elif isinstance(self.actor, NetworkUserAccessActorV1):
            actor = self.actor.to_dict()
        elif isinstance(self.actor, NetworkIPAccessActor):
            actor = self.actor.to_dict()
        else:
            actor = self.actor.to_dict()

        context: Union[None, dict[str, Any]]
        if isinstance(self.context, AccessSessionContextType0):
            context = self.context.to_dict()
        else:
            context = self.context

        created_at = self.created_at.isoformat()

        object_urn = self.object_urn

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "organization_uuid": organization_uuid,
                "actor": actor,
                "context": context,
                "created_at": created_at,
                "object_urn": object_urn,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.access_session_context_type_0 import AccessSessionContextType0
        from ..models.network_ip_access_actor import NetworkIPAccessActor
        from ..models.network_user_access_actor import NetworkUserAccessActor
        from ..models.network_user_access_actor_v1 import NetworkUserAccessActorV1
        from ..models.system_access_actor import SystemAccessActor

        d = dict(src_dict)
        organization_uuid = UUID(d.pop("organization_uuid"))

        def _parse_actor(
            data: object,
        ) -> Union[
            "NetworkIPAccessActor",
            "NetworkUserAccessActor",
            "NetworkUserAccessActorV1",
            "SystemAccessActor",
        ]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                actor_type_0 = NetworkUserAccessActor.from_dict(data)

                return actor_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                actor_type_1 = NetworkUserAccessActorV1.from_dict(data)

                return actor_type_1
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                actor_type_2 = NetworkIPAccessActor.from_dict(data)

                return actor_type_2
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            actor_type_3 = SystemAccessActor.from_dict(data)

            return actor_type_3

        actor = _parse_actor(d.pop("actor"))

        def _parse_context(data: object) -> Union["AccessSessionContextType0", None]:
            if data is None:
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                context_type_0 = AccessSessionContextType0.from_dict(data)

                return context_type_0
            except:  # noqa: E722
                pass
            return cast(Union["AccessSessionContextType0", None], data)

        context = _parse_context(d.pop("context"))

        created_at = isoparse(d.pop("created_at"))

        object_urn = d.pop("object_urn")

        access_session = cls(
            organization_uuid=organization_uuid,
            actor=actor,
            context=context,
            created_at=created_at,
            object_urn=object_urn,
        )

        access_session.additional_properties = d
        return access_session

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.resource_set import ResourceSet


T = TypeVar("T", bound="ResourceSetWithExtras")


@_attrs_define
class ResourceSetWithExtras:
    """
    Attributes:
        resource_set (ResourceSet):
        attached_networks (list[str]):
    """

    resource_set: "ResourceSet"
    attached_networks: list[str]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        resource_set = self.resource_set.to_dict()

        attached_networks = self.attached_networks

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "resource_set": resource_set,
                "attached_networks": attached_networks,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.resource_set import ResourceSet

        d = dict(src_dict)
        resource_set = ResourceSet.from_dict(d.pop("resource_set"))

        attached_networks = cast(list[str], d.pop("attached_networks"))

        resource_set_with_extras = cls(
            resource_set=resource_set,
            attached_networks=attached_networks,
        )

        resource_set_with_extras.additional_properties = d
        return resource_set_with_extras

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.document_type import DocumentType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.editable_document_metadata_index_type_0 import (
        EditableDocumentMetadataIndexType0,
    )


T = TypeVar("T", bound="EditableDocumentMetadata")


@_attrs_define
class EditableDocumentMetadata:
    """
    Attributes:
        name (str):
        document_type (DocumentType):
        index (Union['EditableDocumentMetadataIndexType0', None, Unset]):
    """

    name: str
    document_type: DocumentType
    index: Union["EditableDocumentMetadataIndexType0", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.editable_document_metadata_index_type_0 import (
            EditableDocumentMetadataIndexType0,
        )

        name = self.name

        document_type = self.document_type.value

        index: Union[None, Unset, dict[str, Any]]
        if isinstance(self.index, Unset):
            index = UNSET
        elif isinstance(self.index, EditableDocumentMetadataIndexType0):
            index = self.index.to_dict()
        else:
            index = self.index

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "document_type": document_type,
            }
        )
        if index is not UNSET:
            field_dict["index"] = index

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.editable_document_metadata_index_type_0 import (
            EditableDocumentMetadataIndexType0,
        )

        d = dict(src_dict)
        name = d.pop("name")

        document_type = DocumentType(d.pop("document_type"))

        def _parse_index(
            data: object,
        ) -> Union["EditableDocumentMetadataIndexType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                index_type_0 = EditableDocumentMetadataIndexType0.from_dict(data)

                return index_type_0
            except:  # noqa: E722
                pass
            return cast(Union["EditableDocumentMetadataIndexType0", None, Unset], data)

        index = _parse_index(d.pop("index", UNSET))

        editable_document_metadata = cls(
            name=name,
            document_type=document_type,
            index=index,
        )

        editable_document_metadata.additional_properties = d
        return editable_document_metadata

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union, cast

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.base_error_response import BaseErrorResponse
from ...models.body_upload_snapshot_organization_name_api_v1_uploadsnapshot_post import (
    BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost,
)
from ...models.challenge_response import ChallengeResponse
from ...models.upload_snapshot_response import UploadSnapshotResponse
from ...models.validation_error_response import ValidationErrorResponse
from ...types import UNSET, Response, Unset


def _get_kwargs(
    organization_name: str,
    *,
    body: BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost,
    network: Union[None, Unset, str] = UNSET,
    role: Union[None, Unset, str] = UNSET,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    params: dict[str, Any] = {}

    json_network: Union[None, Unset, str]
    if isinstance(network, Unset):
        json_network = UNSET
    else:
        json_network = network
    params["network"] = json_network

    json_role: Union[None, Unset, str]
    if isinstance(role, Unset):
        json_role = UNSET
    else:
        json_role = role
    params["role"] = json_role

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "post",
        "url": "/{organization_name}/api/v1/uploadsnapshot/".format(
            organization_name=organization_name,
        ),
        "params": params,
    }

    _body = body.to_multipart()

    _kwargs["files"] = _body

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[
        Any,
        BaseErrorResponse,
        ChallengeResponse,
        UploadSnapshotResponse,
        ValidationErrorResponse,
    ]
]:
    if response.status_code == 200:
        response_200 = UploadSnapshotResponse.from_dict(response.json())

        return response_200
    if response.status_code == 422:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == 401:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if response.status_code == 404:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if response.status_code == 413:
        response_413 = cast(Any, None)
        return response_413
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[
        Any,
        BaseErrorResponse,
        ChallengeResponse,
        UploadSnapshotResponse,
        ValidationErrorResponse,
    ]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost,
    network: Union[None, Unset, str] = UNSET,
    role: Union[None, Unset, str] = UNSET,
) -> Response[
    Union[
        Any,
        BaseErrorResponse,
        ChallengeResponse,
        UploadSnapshotResponse,
        ValidationErrorResponse,
    ]
]:
    """Upload a snapshot

     Create a snapshot by direct upload. This triggers snapshot evaluation and rule processing.

    Args:
        organization_name (str):
        network (Union[None, Unset, str]):
        role (Union[None, Unset, str]):
        body (BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, UploadSnapshotResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        body=body,
        network=network,
        role=role,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost,
    network: Union[None, Unset, str] = UNSET,
    role: Union[None, Unset, str] = UNSET,
) -> Optional[
    Union[
        Any,
        BaseErrorResponse,
        ChallengeResponse,
        UploadSnapshotResponse,
        ValidationErrorResponse,
    ]
]:
    """Upload a snapshot

     Create a snapshot by direct upload. This triggers snapshot evaluation and rule processing.

    Args:
        organization_name (str):
        network (Union[None, Unset, str]):
        role (Union[None, Unset, str]):
        body (BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, UploadSnapshotResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        organization_name=organization_name,
        client=client,
        body=body,
        network=network,
        role=role,
    ).parsed


async def asyncio_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost,
    network: Union[None, Unset, str] = UNSET,
    role: Union[None, Unset, str] = UNSET,
) -> Response[
    Union[
        Any,
        BaseErrorResponse,
        ChallengeResponse,
        UploadSnapshotResponse,
        ValidationErrorResponse,
    ]
]:
    """Upload a snapshot

     Create a snapshot by direct upload. This triggers snapshot evaluation and rule processing.

    Args:
        organization_name (str):
        network (Union[None, Unset, str]):
        role (Union[None, Unset, str]):
        body (BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, UploadSnapshotResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        body=body,
        network=network,
        role=role,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost,
    network: Union[None, Unset, str] = UNSET,
    role: Union[None, Unset, str] = UNSET,
) -> Optional[
    Union[
        Any,
        BaseErrorResponse,
        ChallengeResponse,
        UploadSnapshotResponse,
        ValidationErrorResponse,
    ]
]:
    """Upload a snapshot

     Create a snapshot by direct upload. This triggers snapshot evaluation and rule processing.

    Args:
        organization_name (str):
        network (Union[None, Unset, str]):
        role (Union[None, Unset, str]):
        body (BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, UploadSnapshotResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            organization_name=organization_name,
            client=client,
            body=body,
            network=network,
            role=role,
        )
    ).parsed



import argparse
import datetime
import json
import sys
import uuid
import typing
from attrs import asdict
from rich import print_json
from invariant_client import display
from invariant_client.base_command.base_command import BaseCommand
from invariant_client.pysdk import OutputFormat, RemoteError

if typing.TYPE_CHECKING:
    import argparse

def serialize_networks(inst, field, value):
    if isinstance(value, datetime.datetime):
        return value.isoformat()
    if isinstance(value, uuid.UUID):
        return str(value)
    return value

class NetworkCommand(BaseCommand):
    use_argument_debug = True
    use_argument_group_format = True
    use_argument_format_tsv = True

    target_network: str | None = None

    def __init__(self):
        super().__init__()
        self.network_subcommand: str | None = None
        self.network_name: str | None = None
        self.new_network_name: str | None = None
        self.delete_yes: bool = False

    @classmethod
    def parse_args(cls, subparsers: 'argparse._SubParsersAction[argparse.ArgumentParser]') -> None:
        command_networks = subparsers.add_parser(
            'networks',
            description='List or manage networks.',
            help='List or manage networks.'
        )
        cls._add_common_parser_arguments(command_networks)

        networks_subparsers = command_networks.add_subparsers(
            dest='network_subcommand',
            required=False,
            help='Available subcommands for networks'
        )

        list_parser = networks_subparsers.add_parser(
            'list',
            help='List all networks (default action if no subcommand is given).'
        )
        cls._add_common_parser_arguments(list_parser)

        create_parser = networks_subparsers.add_parser(
            'create',
            help='Create a new network.'
        )
        create_parser.add_argument(
            'network_name',
            help='Name of the network to create.'
        )
        cls._add_common_parser_arguments(create_parser)

        rename_parser = networks_subparsers.add_parser(
            'rename',
            help='Rename an existing network.'
        )
        rename_parser.add_argument(
            'network_name',
            help='Current name of the network.'
        )
        rename_parser.add_argument(
            '--new-name',
            required=True,
            dest='new_name',
            help='New name for the network.'
        )
        cls._add_common_parser_arguments(rename_parser)

        delete_parser = networks_subparsers.add_parser(
            'delete',
            help='Delete a network and all its contents (e.g., snapshots).'
        )
        delete_parser.add_argument(
            'network_name',
            help='Name of the network to delete.'
        )
        delete_parser.add_argument(
            '--yes',
            action='store_true',
            help='Confirm deletion without prompting.'
        )
        cls._add_common_parser_arguments(delete_parser)
        
        command_networks.set_defaults(network_subcommand='list')

    def set_config(self, args: 'argparse.Namespace', env: dict[str, str]) -> None:
        super().set_config(args, env)
        self.network_subcommand = getattr(args, 'network_subcommand', 'list')
        self.network_name = getattr(args, 'network_name', None)
        self.new_network_name = getattr(args, 'new_name', None)
        self.delete_yes = getattr(args, 'yes', False)

    def execute(self) -> None:
        super().execute()

        if self.network_subcommand == 'list':
            networks_response = self.sdk.list_networks()
            networks = [asdict(net, value_serializer=serialize_networks) for net in networks_response.networks]
            if self.format == OutputFormat.JSON:
                print_json(data=networks, default=vars)
            elif self.format == OutputFormat.FAST_JSON:
                print(json.dumps(networks, default=vars))
            else:
                display_table = []
                for net_data in networks:
                    row = {
                        "name": net_data.get("name"),
                        "uuid": net_data.get("uuid"),
                        "comment": net_data.get("comment"),
                        "created_at": net_data.get("created_at"),
                    }
                    display_table.append(row)
                display.print_frame(display_table, self.format)

        elif self.network_subcommand == 'create':
            if not self.network_name:
                raise ValueError("Network name required.")
            self.sdk.create_network(name=self.network_name)
            print(f"Network {self.network_name} created.")

        elif self.network_subcommand == 'rename':
            if not self.network_name or not self.new_network_name:
                raise ValueError("Current network name and new network name are required for rename.")
            self.sdk.rename_network_by_name(network_name=self.network_name, new_name=self.new_network_name)
            print(f"Network {self.network_name} renamed to {self.new_network_name}.")

        elif self.network_subcommand == 'delete':
            if not self.network_name:
                raise ValueError("Network name is required for delete.")
            networks_response = self.sdk.list_networks()
            network_to_delete = None
            for net in networks_response.networks:
                if net.name == self.network_name:
                    network_to_delete = net
                    break
            if not network_to_delete or not network_to_delete.uuid:
                print(f"Network {self.network_name} not found.", file=sys.stderr)
                exit(1)
            num_snapshots = 0
            try:
                snapshots = self.sdk.list_snapshots(filter_net=self.network_name)
                num_snapshots = len(snapshots)
            except RemoteError as e:
                if self.debug:
                    print(f"Could not retrieve snapshots for network {self.network_name}: {e}", file=sys.stderr)
            confirmation_message = f"This will delete network {self.network_name}"
            if num_snapshots > 0:
                confirmation_message += f" and {num_snapshots} snapshot{'s' if num_snapshots > 1 else ''}."
            else:
                confirmation_message += "."
            if not self.delete_yes:
                print(confirmation_message)
                confirm = input("Are you sure (y/n)? ")
                if confirm.lower() != 'y':
                    print("Deletion cancelled.")
                    return
            self.sdk.delete_network(network_uuid=network_to_delete.uuid)
            print(f"Network {self.network_name} deleted.")



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.usage_data import UsageData
    from ..models.usage_metadata import UsageMetadata


T = TypeVar("T", bound="UsageModel")


@_attrs_define
class UsageModel:
    """
    Attributes:
        organization_uuid (UUID):
        network_uuid (UUID):
        exec_uuid (UUID):
        data (UsageData):
        metadata (UsageMetadata):
        created_at (Union[Unset, datetime.datetime]):
    """

    organization_uuid: UUID
    network_uuid: UUID
    exec_uuid: UUID
    data: "UsageData"
    metadata: "UsageMetadata"
    created_at: Union[Unset, datetime.datetime] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        organization_uuid = str(self.organization_uuid)

        network_uuid = str(self.network_uuid)

        exec_uuid = str(self.exec_uuid)

        data = self.data.to_dict()

        metadata = self.metadata.to_dict()

        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "organization_uuid": organization_uuid,
                "network_uuid": network_uuid,
                "exec_uuid": exec_uuid,
                "data": data,
                "metadata": metadata,
            }
        )
        if created_at is not UNSET:
            field_dict["created_at"] = created_at

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.usage_data import UsageData
        from ..models.usage_metadata import UsageMetadata

        d = dict(src_dict)
        organization_uuid = UUID(d.pop("organization_uuid"))

        network_uuid = UUID(d.pop("network_uuid"))

        exec_uuid = UUID(d.pop("exec_uuid"))

        data = UsageData.from_dict(d.pop("data"))

        metadata = UsageMetadata.from_dict(d.pop("metadata"))

        _created_at = d.pop("created_at", UNSET)
        created_at: Union[Unset, datetime.datetime]
        if isinstance(_created_at, Unset):
            created_at = UNSET
        else:
            created_at = isoparse(_created_at)

        usage_model = cls(
            organization_uuid=organization_uuid,
            network_uuid=network_uuid,
            exec_uuid=exec_uuid,
            data=data,
            metadata=metadata,
            created_at=created_at,
        )

        usage_model.additional_properties = d
        return usage_model

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="ErrorInfoExtras")


@_attrs_define
class ErrorInfoExtras:
    """Extra data for this error instance. Clients should largely display this data verbatim.

    Attributes:
        file (Union[None, Unset, str]):
        start (Union[None, Unset, list[int]]):
        end (Union[None, Unset, list[int]]):
    """

    file: Union[None, Unset, str] = UNSET
    start: Union[None, Unset, list[int]] = UNSET
    end: Union[None, Unset, list[int]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        file: Union[None, Unset, str]
        if isinstance(self.file, Unset):
            file = UNSET
        else:
            file = self.file

        start: Union[None, Unset, list[int]]
        if isinstance(self.start, Unset):
            start = UNSET
        elif isinstance(self.start, list):
            start = []
            for start_type_0_item_data in self.start:
                start_type_0_item: int
                start_type_0_item = start_type_0_item_data
                start.append(start_type_0_item)

        else:
            start = self.start

        end: Union[None, Unset, list[int]]
        if isinstance(self.end, Unset):
            end = UNSET
        elif isinstance(self.end, list):
            end = []
            for end_type_0_item_data in self.end:
                end_type_0_item: int
                end_type_0_item = end_type_0_item_data
                end.append(end_type_0_item)

        else:
            end = self.end

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if file is not UNSET:
            field_dict["file"] = file
        if start is not UNSET:
            field_dict["start"] = start
        if end is not UNSET:
            field_dict["end"] = end

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)

        def _parse_file(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        file = _parse_file(d.pop("file", UNSET))

        def _parse_start(data: object) -> Union[None, Unset, list[int]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                start_type_0 = []
                _start_type_0 = data
                for start_type_0_item_data in _start_type_0:

                    def _parse_start_type_0_item(data: object) -> int:
                        return cast(int, data)

                    start_type_0_item = _parse_start_type_0_item(start_type_0_item_data)

                    start_type_0.append(start_type_0_item)

                return start_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[int]], data)

        start = _parse_start(d.pop("start", UNSET))

        def _parse_end(data: object) -> Union[None, Unset, list[int]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                end_type_0 = []
                _end_type_0 = data
                for end_type_0_item_data in _end_type_0:

                    def _parse_end_type_0_item(data: object) -> int:
                        return cast(int, data)

                    end_type_0_item = _parse_end_type_0_item(end_type_0_item_data)

                    end_type_0.append(end_type_0_item)

                return end_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[int]], data)

        end = _parse_end(d.pop("end", UNSET))

        error_info_extras = cls(
            file=file,
            start=start,
            end=end,
        )

        error_info_extras.additional_properties = d
        return error_info_extras

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.usage_exec_type import UsageExecType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.access_session import AccessSession


T = TypeVar("T", bound="UsageMetadata")


@_attrs_define
class UsageMetadata:
    """
    Attributes:
        exec_type (UsageExecType): Tracks the source execution type.
        access_session (AccessSession): Represents a data acess actor. This could represent an end user accessing data
            over the network or some internal system action.

        snapshot_uuid (Union[None, UUID, Unset]):
    """

    exec_type: UsageExecType
    access_session: "AccessSession"
    snapshot_uuid: Union[None, UUID, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        exec_type = self.exec_type.value

        access_session = self.access_session.to_dict()

        snapshot_uuid: Union[None, Unset, str]
        if isinstance(self.snapshot_uuid, Unset):
            snapshot_uuid = UNSET
        elif isinstance(self.snapshot_uuid, UUID):
            snapshot_uuid = str(self.snapshot_uuid)
        else:
            snapshot_uuid = self.snapshot_uuid

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "exec_type": exec_type,
                "access_session": access_session,
            }
        )
        if snapshot_uuid is not UNSET:
            field_dict["snapshot_uuid"] = snapshot_uuid

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.access_session import AccessSession

        d = dict(src_dict)
        exec_type = UsageExecType(d.pop("exec_type"))

        access_session = AccessSession.from_dict(d.pop("access_session"))

        def _parse_snapshot_uuid(data: object) -> Union[None, UUID, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                snapshot_uuid_type_0 = UUID(data)

                return snapshot_uuid_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, UUID, Unset], data)

        snapshot_uuid = _parse_snapshot_uuid(d.pop("snapshot_uuid", UNSET))

        usage_metadata = cls(
            exec_type=exec_type,
            access_session=access_session,
            snapshot_uuid=snapshot_uuid,
        )

        usage_metadata.additional_properties = d
        return usage_metadata

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Contains all the data models used in inputs/outputs"""

from .access_session import AccessSession
from .access_session_context_type_0 import AccessSessionContextType0
from .api_token import APIToken
from .api_token_metadata import APITokenMetadata
from .api_token_response import APITokenResponse
from .api_token_response_users import APITokenResponseUsers
from .authn_challenge import AuthnChallenge
from .base_error_response import BaseErrorResponse
from .basic_auth_login_method import BasicAuthLoginMethod
from .body_import_editable_document_organization_name_api_v1_edoc_import_post import (
    BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost,
)
from .body_upload_snapshot_organization_name_api_v1_uploadsnapshot_post import (
    BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost,
)
from .built_in_role import BuiltInRole
from .challenge_response import ChallengeResponse
from .comparison_reportdata import ComparisonReportdata
from .comparison_reportdata_files import ComparisonReportdataFiles
from .console_request_options import ConsoleRequestOptions
from .create_editable_document_request import CreateEditableDocumentRequest
from .create_integration_request_github_app_installation import (
    CreateIntegrationRequestGithubAppInstallation,
)
from .create_integration_request_github_app_installation_data import (
    CreateIntegrationRequestGithubAppInstallationData,
)
from .create_integration_request_slack_app_installation import (
    CreateIntegrationRequestSlackAppInstallation,
)
from .create_integration_request_slack_app_installation_data import (
    CreateIntegrationRequestSlackAppInstallationData,
)
from .create_managed_user_request import CreateManagedUserRequest
from .create_member_response import CreateMemberResponse
from .create_monitor_target_request import CreateMonitorTargetRequest
from .create_network_request import CreateNetworkRequest
from .create_notification_group_request import CreateNotificationGroupRequest
from .create_resource_set_request import CreateResourceSetRequest
from .create_security_integration_request import CreateSecurityIntegrationRequest
from .create_service_account_request import CreateServiceAccountRequest
from .create_service_account_response import CreateServiceAccountResponse
from .create_token_request import CreateTokenRequest
from .credit_stats import CreditStats
from .document_type import DocumentType
from .editable_document import EditableDocument
from .editable_document_metadata import EditableDocumentMetadata
from .editable_document_metadata_index_type_0 import EditableDocumentMetadataIndexType0
from .editable_document_result import EditableDocumentResult
from .editable_document_version import EditableDocumentVersion
from .editable_document_version_metadata import EditableDocumentVersionMetadata
from .editable_document_version_metadata_index_type_0 import (
    EditableDocumentVersionMetadataIndexType0,
)
from .email_subscriber import EmailSubscriber
from .error_info import ErrorInfo
from .error_info_extras import ErrorInfoExtras
from .exec_response import ExecResponse
from .exec_response_result_files import ExecResponseResultFiles
from .exec_response_results import ExecResponseResults
from .external_status_data_integration import ExternalStatusDataIntegration
from .external_status_integration import ExternalStatusIntegration
from .file_index import FileIndex
from .flags_response import FlagsResponse
from .flags_response_environment import FlagsResponseEnvironment
from .flags_response_flags import FlagsResponseFlags
from .generic_state import GenericState
from .get_report_summary_response import GetReportSummaryResponse
from .get_report_summary_response_status import GetReportSummaryResponseStatus
from .get_report_summary_response_summary import GetReportSummaryResponseSummary
from .get_resource_set_response import GetResourceSetResponse
from .get_resource_set_response_with_extras import GetResourceSetResponseWithExtras
from .get_usage_stats_organization_name_api_v1_usage_stats_get_period import (
    GetUsageStatsOrganizationNameApiV1UsageStatsGetPeriod,
)
from .github_branch import GithubBranch
from .github_commit import GithubCommit
from .github_repository import GithubRepository
from .github_repository_data import GithubRepositoryData
from .integration import Integration
from .integration_data_github_app_installation import (
    IntegrationDataGithubAppInstallation,
)
from .integration_data_github_app_installation_data import (
    IntegrationDataGithubAppInstallationData,
)
from .integration_data_github_app_installation_data_extra import (
    IntegrationDataGithubAppInstallationDataExtra,
)
from .integration_data_slack_app_installation import IntegrationDataSlackAppInstallation
from .integration_data_slack_app_installation_data import (
    IntegrationDataSlackAppInstallationData,
)
from .integration_with_status_github_installation import (
    IntegrationWithStatusGithubInstallation,
)
from .integration_with_status_slack_app import IntegrationWithStatusSlackApp
from .invariant_plan import InvariantPlan
from .invariant_plan_response import InvariantPlanResponse
from .invariant_price import InvariantPrice
from .invite_response import InviteResponse
from .invite_user_request import InviteUserRequest
from .list_networks_response import ListNetworksResponse
from .list_networks_with_extras_response import ListNetworksWithExtrasResponse
from .list_notification_groups_response import ListNotificationGroupsResponse
from .list_report_tasks_response import ListReportTasksResponse
from .list_reports_response import ListReportsResponse
from .list_snapshots_response import ListSnapshotsResponse
from .list_usage_response import ListUsageResponse
from .login_config_metadata_public import LoginConfigMetadataPublic
from .login_config_public import LoginConfigPublic
from .modify_allow_inbound_invitations_request import (
    ModifyAllowInboundInvitationsRequest,
)
from .modify_allow_outbound_invitations_request import (
    ModifyAllowOutboundInvitationsRequest,
)
from .modify_default_login_methods_request import ModifyDefaultLoginMethodsRequest
from .modify_network_request import ModifyNetworkRequest
from .modify_user_request import ModifyUserRequest
from .monitor_target import MonitorTarget
from .monitor_target_metadata import MonitorTargetMetadata
from .network import Network
from .network_ip_access_actor import NetworkIPAccessActor
from .network_metadata import NetworkMetadata
from .network_resource_set_attachment_request import NetworkResourceSetAttachmentRequest
from .network_user_access_actor import NetworkUserAccessActor
from .network_user_access_actor_v1 import NetworkUserAccessActorV1
from .network_with_extras import NetworkWithExtras
from .new_login_challenge import NewLoginChallenge
from .notification_group import NotificationGroup
from .notification_group_metadata import NotificationGroupMetadata
from .oidc_login_method import OIDCLoginMethod
from .oidc_principal import OIDCPrincipal
from .oidc_security_integration_metadata import OIDCSecurityIntegrationMetadata
from .org_metadata import OrgMetadata
from .organization import Organization
from .organization_member_with_extras import OrganizationMemberWithExtras
from .password_reset_pin_challenge import PasswordResetPINChallenge
from .poc_report_data import POCReportData
from .policy_file_result import PolicyFileResult
from .policy_file_result_outcome import PolicyFileResultOutcome
from .policy_file_result_policy import PolicyFileResultPolicy
from .policy_file_result_rule import PolicyFileResultRule
from .public import Public
from .refresh_response import RefreshResponse
from .report import Report
from .report_extras import ReportExtras
from .report_metadata import ReportMetadata
from .report_task import ReportTask
from .report_text_summary_request import ReportTextSummaryRequest
from .report_text_summary_response import ReportTextSummaryResponse
from .repository import Repository
from .resource_set import ResourceSet
from .resource_set_member import ResourceSetMember
from .resource_set_metadata import ResourceSetMetadata
from .resource_set_with_extras import ResourceSetWithExtras
from .resource_type import ResourceType
from .rule_outcome import RuleOutcome
from .rule_summary_group_uploaded import RuleSummaryGroupUploaded
from .rule_summary_group_webrule import RuleSummaryGroupWebrule
from .rule_summary_stats import RuleSummaryStats
from .rule_type import RuleType
from .security_integration import SecurityIntegration
from .security_policy_metadata import SecurityPolicyMetadata
from .security_settings_response import SecuritySettingsResponse
from .set_password_challenge import SetPasswordChallenge
from .setup_checkout_expansion_request import SetupCheckoutExpansionRequest
from .setup_checkout_request import SetupCheckoutRequest
from .setup_checkout_response import SetupCheckoutResponse
from .setup_code_challenge import SetupCodeChallenge
from .slack_channel import SlackChannel
from .slack_integration import SlackIntegration
from .slack_metadata import SlackMetadata
from .slack_subscriber import SlackSubscriber
from .snapshot_metadata import SnapshotMetadata
from .snapshot_model import SnapshotModel
from .snapshot_report_data import SnapshotReportData
from .snapshot_report_data_files import SnapshotReportDataFiles
from .start_challenge import StartChallenge
from .stripe_line_items_data import StripeLineItemsData
from .stripe_line_items_response import StripeLineItemsResponse
from .stripe_line_items_response_item_data import StripeLineItemsResponseItemData
from .stripe_products_response import StripeProductsResponse
from .stripe_products_response_plans import StripeProductsResponsePlans
from .stripe_transaction_status_response import StripeTransactionStatusResponse
from .stripe_transaction_status_response_status import (
    StripeTransactionStatusResponseStatus,
)
from .subscription_line_item import SubscriptionLineItem
from .system_access_actor import SystemAccessActor
from .tab_info import TabInfo
from .tab_info_parameters_type_0 import TabInfoParametersType0
from .tab_info_state_type_0 import TabInfoStateType0
from .test_notification_group_request import TestNotificationGroupRequest
from .try_rule_request import TryRuleRequest
from .ui_status_response import UIStatusResponse
from .ui_status_response_permissions import UIStatusResponsePermissions
from .update_editable_document_conflict_error_response import (
    UpdateEditableDocumentConflictErrorResponse,
)
from .update_editable_document_name_update import UpdateEditableDocumentNameUpdate
from .update_editable_document_request import UpdateEditableDocumentRequest
from .update_snapshots_request import UpdateSnapshotsRequest
from .update_snapshots_response import UpdateSnapshotsResponse
from .updated_snapshot_detail import UpdatedSnapshotDetail
from .upload_snapshot_response import UploadSnapshotResponse
from .upload_snapshot_status_response import UploadSnapshotStatusResponse
from .usage_data import UsageData
from .usage_data_inventory import UsageDataInventory
from .usage_exec_type import UsageExecType
from .usage_metadata import UsageMetadata
from .usage_model import UsageModel
from .usage_stats import UsageStats
from .usage_stats_response import UsageStatsResponse
from .user import User
from .user_metadata import UserMetadata
from .user_tabs_config import UserTabsConfig
from .uuid_volume_locator import UUIDVolumeLocator
from .validate_email_challenge import ValidateEmailChallenge
from .validation_error_response import ValidationErrorResponse
from .validation_error_response_part import ValidationErrorResponsePart

__all__ = (
    "AccessSession",
    "AccessSessionContextType0",
    "APIToken",
    "APITokenMetadata",
    "APITokenResponse",
    "APITokenResponseUsers",
    "AuthnChallenge",
    "BaseErrorResponse",
    "BasicAuthLoginMethod",
    "BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost",
    "BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost",
    "BuiltInRole",
    "ChallengeResponse",
    "ComparisonReportdata",
    "ComparisonReportdataFiles",
    "ConsoleRequestOptions",
    "CreateEditableDocumentRequest",
    "CreateIntegrationRequestGithubAppInstallation",
    "CreateIntegrationRequestGithubAppInstallationData",
    "CreateIntegrationRequestSlackAppInstallation",
    "CreateIntegrationRequestSlackAppInstallationData",
    "CreateManagedUserRequest",
    "CreateMemberResponse",
    "CreateMonitorTargetRequest",
    "CreateNetworkRequest",
    "CreateNotificationGroupRequest",
    "CreateResourceSetRequest",
    "CreateSecurityIntegrationRequest",
    "CreateServiceAccountRequest",
    "CreateServiceAccountResponse",
    "CreateTokenRequest",
    "CreditStats",
    "DocumentType",
    "EditableDocument",
    "EditableDocumentMetadata",
    "EditableDocumentMetadataIndexType0",
    "EditableDocumentResult",
    "EditableDocumentVersion",
    "EditableDocumentVersionMetadata",
    "EditableDocumentVersionMetadataIndexType0",
    "EmailSubscriber",
    "ErrorInfo",
    "ErrorInfoExtras",
    "ExecResponse",
    "ExecResponseResultFiles",
    "ExecResponseResults",
    "ExternalStatusDataIntegration",
    "ExternalStatusIntegration",
    "FileIndex",
    "FlagsResponse",
    "FlagsResponseEnvironment",
    "FlagsResponseFlags",
    "GenericState",
    "GetReportSummaryResponse",
    "GetReportSummaryResponseStatus",
    "GetReportSummaryResponseSummary",
    "GetResourceSetResponse",
    "GetResourceSetResponseWithExtras",
    "GetUsageStatsOrganizationNameApiV1UsageStatsGetPeriod",
    "GithubBranch",
    "GithubCommit",
    "GithubRepository",
    "GithubRepositoryData",
    "Integration",
    "IntegrationDataGithubAppInstallation",
    "IntegrationDataGithubAppInstallationData",
    "IntegrationDataGithubAppInstallationDataExtra",
    "IntegrationDataSlackAppInstallation",
    "IntegrationDataSlackAppInstallationData",
    "IntegrationWithStatusGithubInstallation",
    "IntegrationWithStatusSlackApp",
    "InvariantPlan",
    "InvariantPlanResponse",
    "InvariantPrice",
    "InviteResponse",
    "InviteUserRequest",
    "ListNetworksResponse",
    "ListNetworksWithExtrasResponse",
    "ListNotificationGroupsResponse",
    "ListReportsResponse",
    "ListReportTasksResponse",
    "ListSnapshotsResponse",
    "ListUsageResponse",
    "LoginConfigMetadataPublic",
    "LoginConfigPublic",
    "ModifyAllowInboundInvitationsRequest",
    "ModifyAllowOutboundInvitationsRequest",
    "ModifyDefaultLoginMethodsRequest",
    "ModifyNetworkRequest",
    "ModifyUserRequest",
    "MonitorTarget",
    "MonitorTargetMetadata",
    "Network",
    "NetworkIPAccessActor",
    "NetworkMetadata",
    "NetworkResourceSetAttachmentRequest",
    "NetworkUserAccessActor",
    "NetworkUserAccessActorV1",
    "NetworkWithExtras",
    "NewLoginChallenge",
    "NotificationGroup",
    "NotificationGroupMetadata",
    "OIDCLoginMethod",
    "OIDCPrincipal",
    "OIDCSecurityIntegrationMetadata",
    "Organization",
    "OrganizationMemberWithExtras",
    "OrgMetadata",
    "PasswordResetPINChallenge",
    "POCReportData",
    "PolicyFileResult",
    "PolicyFileResultOutcome",
    "PolicyFileResultPolicy",
    "PolicyFileResultRule",
    "Public",
    "RefreshResponse",
    "Report",
    "ReportExtras",
    "ReportMetadata",
    "ReportTask",
    "ReportTextSummaryRequest",
    "ReportTextSummaryResponse",
    "Repository",
    "ResourceSet",
    "ResourceSetMember",
    "ResourceSetMetadata",
    "ResourceSetWithExtras",
    "ResourceType",
    "RuleOutcome",
    "RuleSummaryGroupUploaded",
    "RuleSummaryGroupWebrule",
    "RuleSummaryStats",
    "RuleType",
    "SecurityIntegration",
    "SecurityPolicyMetadata",
    "SecuritySettingsResponse",
    "SetPasswordChallenge",
    "SetupCheckoutExpansionRequest",
    "SetupCheckoutRequest",
    "SetupCheckoutResponse",
    "SetupCodeChallenge",
    "SlackChannel",
    "SlackIntegration",
    "SlackMetadata",
    "SlackSubscriber",
    "SnapshotMetadata",
    "SnapshotModel",
    "SnapshotReportData",
    "SnapshotReportDataFiles",
    "StartChallenge",
    "StripeLineItemsData",
    "StripeLineItemsResponse",
    "StripeLineItemsResponseItemData",
    "StripeProductsResponse",
    "StripeProductsResponsePlans",
    "StripeTransactionStatusResponse",
    "StripeTransactionStatusResponseStatus",
    "SubscriptionLineItem",
    "SystemAccessActor",
    "TabInfo",
    "TabInfoParametersType0",
    "TabInfoStateType0",
    "TestNotificationGroupRequest",
    "TryRuleRequest",
    "UIStatusResponse",
    "UIStatusResponsePermissions",
    "UpdatedSnapshotDetail",
    "UpdateEditableDocumentConflictErrorResponse",
    "UpdateEditableDocumentNameUpdate",
    "UpdateEditableDocumentRequest",
    "UpdateSnapshotsRequest",
    "UpdateSnapshotsResponse",
    "UploadSnapshotResponse",
    "UploadSnapshotStatusResponse",
    "UsageData",
    "UsageDataInventory",
    "UsageExecType",
    "UsageMetadata",
    "UsageModel",
    "UsageStats",
    "UsageStatsResponse",
    "User",
    "UserMetadata",
    "UserTabsConfig",
    "UUIDVolumeLocator",
    "ValidateEmailChallenge",
    "ValidationErrorResponse",
    "ValidationErrorResponsePart",
)



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.policy_file_result_outcome import PolicyFileResultOutcome


T = TypeVar("T", bound="PolicyFileResult")


@_attrs_define
class PolicyFileResult:
    """
    Attributes:
        uuid (UUID):
        organization_uuid (UUID):
        network_uuid (UUID):
        snapshot_uuid (UUID):
        outcome (PolicyFileResultOutcome):
        created_at (datetime.datetime):
        rule_resource_set_uuid (Union[None, UUID, Unset]):
        rule_editable_document_uuid (Union[None, UUID, Unset]):
        original_filename (Union[None, Unset, str]):
        deleted_at (Union[None, Unset, datetime.datetime]):
    """

    uuid: UUID
    organization_uuid: UUID
    network_uuid: UUID
    snapshot_uuid: UUID
    outcome: "PolicyFileResultOutcome"
    created_at: datetime.datetime
    rule_resource_set_uuid: Union[None, UUID, Unset] = UNSET
    rule_editable_document_uuid: Union[None, UUID, Unset] = UNSET
    original_filename: Union[None, Unset, str] = UNSET
    deleted_at: Union[None, Unset, datetime.datetime] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid = str(self.uuid)

        organization_uuid = str(self.organization_uuid)

        network_uuid = str(self.network_uuid)

        snapshot_uuid = str(self.snapshot_uuid)

        outcome = self.outcome.to_dict()

        created_at = self.created_at.isoformat()

        rule_resource_set_uuid: Union[None, Unset, str]
        if isinstance(self.rule_resource_set_uuid, Unset):
            rule_resource_set_uuid = UNSET
        elif isinstance(self.rule_resource_set_uuid, UUID):
            rule_resource_set_uuid = str(self.rule_resource_set_uuid)
        else:
            rule_resource_set_uuid = self.rule_resource_set_uuid

        rule_editable_document_uuid: Union[None, Unset, str]
        if isinstance(self.rule_editable_document_uuid, Unset):
            rule_editable_document_uuid = UNSET
        elif isinstance(self.rule_editable_document_uuid, UUID):
            rule_editable_document_uuid = str(self.rule_editable_document_uuid)
        else:
            rule_editable_document_uuid = self.rule_editable_document_uuid

        original_filename: Union[None, Unset, str]
        if isinstance(self.original_filename, Unset):
            original_filename = UNSET
        else:
            original_filename = self.original_filename

        deleted_at: Union[None, Unset, str]
        if isinstance(self.deleted_at, Unset):
            deleted_at = UNSET
        elif isinstance(self.deleted_at, datetime.datetime):
            deleted_at = self.deleted_at.isoformat()
        else:
            deleted_at = self.deleted_at

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "uuid": uuid,
                "organization_uuid": organization_uuid,
                "network_uuid": network_uuid,
                "snapshot_uuid": snapshot_uuid,
                "outcome": outcome,
                "created_at": created_at,
            }
        )
        if rule_resource_set_uuid is not UNSET:
            field_dict["rule_resource_set_uuid"] = rule_resource_set_uuid
        if rule_editable_document_uuid is not UNSET:
            field_dict["rule_editable_document_uuid"] = rule_editable_document_uuid
        if original_filename is not UNSET:
            field_dict["original_filename"] = original_filename
        if deleted_at is not UNSET:
            field_dict["deleted_at"] = deleted_at

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.policy_file_result_outcome import PolicyFileResultOutcome

        d = dict(src_dict)
        uuid = UUID(d.pop("uuid"))

        organization_uuid = UUID(d.pop("organization_uuid"))

        network_uuid = UUID(d.pop("network_uuid"))

        snapshot_uuid = UUID(d.pop("snapshot_uuid"))

        outcome = PolicyFileResultOutcome.from_dict(d.pop("outcome"))

        created_at = isoparse(d.pop("created_at"))

        def _parse_rule_resource_set_uuid(data: object) -> Union[None, UUID, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                rule_resource_set_uuid_type_0 = UUID(data)

                return rule_resource_set_uuid_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, UUID, Unset], data)

        rule_resource_set_uuid = _parse_rule_resource_set_uuid(
            d.pop("rule_resource_set_uuid", UNSET)
        )

        def _parse_rule_editable_document_uuid(
            data: object,
        ) -> Union[None, UUID, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                rule_editable_document_uuid_type_0 = UUID(data)

                return rule_editable_document_uuid_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, UUID, Unset], data)

        rule_editable_document_uuid = _parse_rule_editable_document_uuid(
            d.pop("rule_editable_document_uuid", UNSET)
        )

        def _parse_original_filename(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        original_filename = _parse_original_filename(d.pop("original_filename", UNSET))

        def _parse_deleted_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                deleted_at_type_0 = isoparse(data)

                return deleted_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        deleted_at = _parse_deleted_at(d.pop("deleted_at", UNSET))

        policy_file_result = cls(
            uuid=uuid,
            organization_uuid=organization_uuid,
            network_uuid=network_uuid,
            snapshot_uuid=snapshot_uuid,
            outcome=outcome,
            created_at=created_at,
            rule_resource_set_uuid=rule_resource_set_uuid,
            rule_editable_document_uuid=rule_editable_document_uuid,
            original_filename=original_filename,
            deleted_at=deleted_at,
        )

        policy_file_result.additional_properties = d
        return policy_file_result

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="POCReportData")


@_attrs_define
class POCReportData:
    """
    Attributes:
        issues (UUID):
        edges (UUID):
        routers (UUID):
        nodes (UUID):
        external_ports (UUID):
        rule_findings (UUID):
        connect_to (UUID):
    """

    issues: UUID
    edges: UUID
    routers: UUID
    nodes: UUID
    external_ports: UUID
    rule_findings: UUID
    connect_to: UUID
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        issues = str(self.issues)

        edges = str(self.edges)

        routers = str(self.routers)

        nodes = str(self.nodes)

        external_ports = str(self.external_ports)

        rule_findings = str(self.rule_findings)

        connect_to = str(self.connect_to)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "issues": issues,
                "edges": edges,
                "routers": routers,
                "nodes": nodes,
                "external_ports": external_ports,
                "rule_findings": rule_findings,
                "connectTo": connect_to,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        issues = UUID(d.pop("issues"))

        edges = UUID(d.pop("edges"))

        routers = UUID(d.pop("routers"))

        nodes = UUID(d.pop("nodes"))

        external_ports = UUID(d.pop("external_ports"))

        rule_findings = UUID(d.pop("rule_findings"))

        connect_to = UUID(d.pop("connectTo"))

        poc_report_data = cls(
            issues=issues,
            edges=edges,
            routers=routers,
            nodes=nodes,
            external_ports=external_ports,
            rule_findings=rule_findings,
            connect_to=connect_to,
        )

        poc_report_data.additional_properties = d
        return poc_report_data

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.base_error_response import BaseErrorResponse
from ...models.challenge_response import ChallengeResponse
from ...models.list_snapshots_response import ListSnapshotsResponse
from ...models.validation_error_response import ValidationErrorResponse
from ...types import UNSET, Response, Unset


def _get_kwargs(
    organization_name: str,
    *,
    network_name: Union[None, Unset, str] = UNSET,
    limit: Union[None, Unset, int] = 0,
    filter_session: Union[Unset, bool] = False,
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    json_network_name: Union[None, Unset, str]
    if isinstance(network_name, Unset):
        json_network_name = UNSET
    else:
        json_network_name = network_name
    params["network_name"] = json_network_name

    json_limit: Union[None, Unset, int]
    if isinstance(limit, Unset):
        json_limit = UNSET
    else:
        json_limit = limit
    params["limit"] = json_limit

    params["filter_session"] = filter_session

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": "/{organization_name}/api/v1/snapshots/".format(
            organization_name=organization_name,
        ),
        "params": params,
    }

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ValidationErrorResponse,
        list["ListSnapshotsResponse"],
    ]
]:
    if response.status_code == 200:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = ListSnapshotsResponse.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    if response.status_code == 422:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == 401:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if response.status_code == 404:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ValidationErrorResponse,
        list["ListSnapshotsResponse"],
    ]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    network_name: Union[None, Unset, str] = UNSET,
    limit: Union[None, Unset, int] = 0,
    filter_session: Union[Unset, bool] = False,
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ValidationErrorResponse,
        list["ListSnapshotsResponse"],
    ]
]:
    """Lists all snapshots for a network. Optionally filters by session.

    Args:
        organization_name (str):
        network_name (Union[None, Unset, str]):
        limit (Union[None, Unset, int]):  Default: 0.
        filter_session (Union[Unset, bool]):  Default: False.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[BaseErrorResponse, ChallengeResponse, ValidationErrorResponse, list['ListSnapshotsResponse']]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        network_name=network_name,
        limit=limit,
        filter_session=filter_session,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    network_name: Union[None, Unset, str] = UNSET,
    limit: Union[None, Unset, int] = 0,
    filter_session: Union[Unset, bool] = False,
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ValidationErrorResponse,
        list["ListSnapshotsResponse"],
    ]
]:
    """Lists all snapshots for a network. Optionally filters by session.

    Args:
        organization_name (str):
        network_name (Union[None, Unset, str]):
        limit (Union[None, Unset, int]):  Default: 0.
        filter_session (Union[Unset, bool]):  Default: False.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[BaseErrorResponse, ChallengeResponse, ValidationErrorResponse, list['ListSnapshotsResponse']]
    """

    return sync_detailed(
        organization_name=organization_name,
        client=client,
        network_name=network_name,
        limit=limit,
        filter_session=filter_session,
    ).parsed


async def asyncio_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    network_name: Union[None, Unset, str] = UNSET,
    limit: Union[None, Unset, int] = 0,
    filter_session: Union[Unset, bool] = False,
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ValidationErrorResponse,
        list["ListSnapshotsResponse"],
    ]
]:
    """Lists all snapshots for a network. Optionally filters by session.

    Args:
        organization_name (str):
        network_name (Union[None, Unset, str]):
        limit (Union[None, Unset, int]):  Default: 0.
        filter_session (Union[Unset, bool]):  Default: False.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[BaseErrorResponse, ChallengeResponse, ValidationErrorResponse, list['ListSnapshotsResponse']]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        network_name=network_name,
        limit=limit,
        filter_session=filter_session,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    network_name: Union[None, Unset, str] = UNSET,
    limit: Union[None, Unset, int] = 0,
    filter_session: Union[Unset, bool] = False,
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ValidationErrorResponse,
        list["ListSnapshotsResponse"],
    ]
]:
    """Lists all snapshots for a network. Optionally filters by session.

    Args:
        organization_name (str):
        network_name (Union[None, Unset, str]):
        limit (Union[None, Unset, int]):  Default: 0.
        filter_session (Union[Unset, bool]):  Default: False.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[BaseErrorResponse, ChallengeResponse, ValidationErrorResponse, list['ListSnapshotsResponse']]
    """

    return (
        await asyncio_detailed(
            organization_name=organization_name,
            client=client,
            network_name=network_name,
            limit=limit,
            filter_session=filter_session,
        )
    ).parsed



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.poc_report_data import POCReportData
    from ..models.report_metadata import ReportMetadata
    from ..models.snapshot_report_data import SnapshotReportData


T = TypeVar("T", bound="Report")


@_attrs_define
class Report:
    """
    Attributes:
        uuid (UUID):
        organization_uuid (UUID):
        network_uuid (UUID):
        reports (Union['POCReportData', 'SnapshotReportData']):
        metadata (ReportMetadata):
        created_at (datetime.datetime):
        snapshot_uuid (Union[None, UUID, Unset]):
    """

    uuid: UUID
    organization_uuid: UUID
    network_uuid: UUID
    reports: Union["POCReportData", "SnapshotReportData"]
    metadata: "ReportMetadata"
    created_at: datetime.datetime
    snapshot_uuid: Union[None, UUID, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.snapshot_report_data import SnapshotReportData

        uuid = str(self.uuid)

        organization_uuid = str(self.organization_uuid)

        network_uuid = str(self.network_uuid)

        reports: dict[str, Any]
        if isinstance(self.reports, SnapshotReportData):
            reports = self.reports.to_dict()
        else:
            reports = self.reports.to_dict()

        metadata = self.metadata.to_dict()

        created_at = self.created_at.isoformat()

        snapshot_uuid: Union[None, Unset, str]
        if isinstance(self.snapshot_uuid, Unset):
            snapshot_uuid = UNSET
        elif isinstance(self.snapshot_uuid, UUID):
            snapshot_uuid = str(self.snapshot_uuid)
        else:
            snapshot_uuid = self.snapshot_uuid

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "uuid": uuid,
                "organization_uuid": organization_uuid,
                "network_uuid": network_uuid,
                "reports": reports,
                "metadata": metadata,
                "created_at": created_at,
            }
        )
        if snapshot_uuid is not UNSET:
            field_dict["snapshot_uuid"] = snapshot_uuid

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.poc_report_data import POCReportData
        from ..models.report_metadata import ReportMetadata
        from ..models.snapshot_report_data import SnapshotReportData

        d = dict(src_dict)
        uuid = UUID(d.pop("uuid"))

        organization_uuid = UUID(d.pop("organization_uuid"))

        network_uuid = UUID(d.pop("network_uuid"))

        def _parse_reports(
            data: object,
        ) -> Union["POCReportData", "SnapshotReportData"]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                reports_type_0 = SnapshotReportData.from_dict(data)

                return reports_type_0
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            reports_type_1 = POCReportData.from_dict(data)

            return reports_type_1

        reports = _parse_reports(d.pop("reports"))

        metadata = ReportMetadata.from_dict(d.pop("metadata"))

        created_at = isoparse(d.pop("created_at"))

        def _parse_snapshot_uuid(data: object) -> Union[None, UUID, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                snapshot_uuid_type_0 = UUID(data)

                return snapshot_uuid_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, UUID, Unset], data)

        snapshot_uuid = _parse_snapshot_uuid(d.pop("snapshot_uuid", UNSET))

        report = cls(
            uuid=uuid,
            organization_uuid=organization_uuid,
            network_uuid=network_uuid,
            reports=reports,
            metadata=metadata,
            created_at=created_at,
            snapshot_uuid=snapshot_uuid,
        )

        report.additional_properties = d
        return report

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.authn_challenge import AuthnChallenge
    from ..models.new_login_challenge import NewLoginChallenge
    from ..models.password_reset_pin_challenge import PasswordResetPINChallenge
    from ..models.set_password_challenge import SetPasswordChallenge
    from ..models.setup_code_challenge import SetupCodeChallenge
    from ..models.start_challenge import StartChallenge
    from ..models.validate_email_challenge import ValidateEmailChallenge


T = TypeVar("T", bound="ChallengeResponse")


@_attrs_define
class ChallengeResponse:
    """
    Attributes:
        status (int):
        type_ (Literal['urn:invariant:errors:auth_challenge']):
        title (str):
        detail (str):
        challenge (Union['AuthnChallenge', 'NewLoginChallenge', 'PasswordResetPINChallenge', 'SetPasswordChallenge',
            'SetupCodeChallenge', 'StartChallenge', 'ValidateEmailChallenge']):
        instance (Union[None, Unset, str]):
        login_token (Union[None, Unset, str]):
    """

    status: int
    type_: Literal["urn:invariant:errors:auth_challenge"]
    title: str
    detail: str
    challenge: Union[
        "AuthnChallenge",
        "NewLoginChallenge",
        "PasswordResetPINChallenge",
        "SetPasswordChallenge",
        "SetupCodeChallenge",
        "StartChallenge",
        "ValidateEmailChallenge",
    ]
    instance: Union[None, Unset, str] = UNSET
    login_token: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.authn_challenge import AuthnChallenge
        from ..models.new_login_challenge import NewLoginChallenge
        from ..models.password_reset_pin_challenge import PasswordResetPINChallenge
        from ..models.set_password_challenge import SetPasswordChallenge
        from ..models.setup_code_challenge import SetupCodeChallenge
        from ..models.validate_email_challenge import ValidateEmailChallenge

        status = self.status

        type_ = self.type_

        title = self.title

        detail = self.detail

        challenge: dict[str, Any]
        if isinstance(self.challenge, AuthnChallenge):
            challenge = self.challenge.to_dict()
        elif isinstance(self.challenge, NewLoginChallenge):
            challenge = self.challenge.to_dict()
        elif isinstance(self.challenge, ValidateEmailChallenge):
            challenge = self.challenge.to_dict()
        elif isinstance(self.challenge, PasswordResetPINChallenge):
            challenge = self.challenge.to_dict()
        elif isinstance(self.challenge, SetPasswordChallenge):
            challenge = self.challenge.to_dict()
        elif isinstance(self.challenge, SetupCodeChallenge):
            challenge = self.challenge.to_dict()
        else:
            challenge = self.challenge.to_dict()

        instance: Union[None, Unset, str]
        if isinstance(self.instance, Unset):
            instance = UNSET
        else:
            instance = self.instance

        login_token: Union[None, Unset, str]
        if isinstance(self.login_token, Unset):
            login_token = UNSET
        else:
            login_token = self.login_token

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "type": type_,
                "title": title,
                "detail": detail,
                "challenge": challenge,
            }
        )
        if instance is not UNSET:
            field_dict["instance"] = instance
        if login_token is not UNSET:
            field_dict["login_token"] = login_token

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.authn_challenge import AuthnChallenge
        from ..models.new_login_challenge import NewLoginChallenge
        from ..models.password_reset_pin_challenge import PasswordResetPINChallenge
        from ..models.set_password_challenge import SetPasswordChallenge
        from ..models.setup_code_challenge import SetupCodeChallenge
        from ..models.start_challenge import StartChallenge
        from ..models.validate_email_challenge import ValidateEmailChallenge

        d = dict(src_dict)
        status = d.pop("status")

        type_ = cast(Literal["urn:invariant:errors:auth_challenge"], d.pop("type"))
        if type_ != "urn:invariant:errors:auth_challenge":
            raise ValueError(
                f"type must match const 'urn:invariant:errors:auth_challenge', got '{type_}'"
            )

        title = d.pop("title")

        detail = d.pop("detail")

        def _parse_challenge(
            data: object,
        ) -> Union[
            "AuthnChallenge",
            "NewLoginChallenge",
            "PasswordResetPINChallenge",
            "SetPasswordChallenge",
            "SetupCodeChallenge",
            "StartChallenge",
            "ValidateEmailChallenge",
        ]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_0 = AuthnChallenge.from_dict(data)

                return challenge_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_1 = NewLoginChallenge.from_dict(data)

                return challenge_type_1
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_2 = ValidateEmailChallenge.from_dict(data)

                return challenge_type_2
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_3 = PasswordResetPINChallenge.from_dict(data)

                return challenge_type_3
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_4 = SetPasswordChallenge.from_dict(data)

                return challenge_type_4
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                challenge_type_5 = SetupCodeChallenge.from_dict(data)

                return challenge_type_5
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            challenge_type_6 = StartChallenge.from_dict(data)

            return challenge_type_6

        challenge = _parse_challenge(d.pop("challenge"))

        def _parse_instance(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        instance = _parse_instance(d.pop("instance", UNSET))

        def _parse_login_token(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        login_token = _parse_login_token(d.pop("login_token", UNSET))

        challenge_response = cls(
            status=status,
            type_=type_,
            title=title,
            detail=detail,
            challenge=challenge,
            instance=instance,
            login_token=login_token,
        )

        challenge_response.additional_properties = d
        return challenge_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
